/*
  James R. Diebel
  Stanford University

  Started: 27 August 2004
  Last revised: 27 Sugust 2004

  vert.cc - class implementation of Vert class (vert.hh)

  Depends on vert.hh
*/

#include "bmtk/vert.hh"

namespace bmtk {

// Constructor/Destructor
Vert::Vert() {
  nf = ne = 0;
  i = -1;
  l = psi = 0.0;
  bound = false;
  flag = marker = -1;
  depth = 0.0f;
}
Vert::~Vert() {
  if (nf) delete [] f;
  if (ne) delete [] e;
  if (ne) delete [] v;
  if (m) if (m->nd) delete [] xs;
}

// Member functions
void Vert::findNormal() {
  n = 0.0f;
  for (int j=0;j<nf;j++) if (f[j]) n += f[j]->n;
  n.normalize();
}

void Vert::findLength() {
  l = 0;
  int count = 0;
  for (int j=0;j<ne;j++) if (e[j]) {l += e[j]->l; count++;}
  if (count > 0) l /= float(count); else l=0;
}

void Vert::findPotential() {
  Vec3d temp = (x - x0);
  psi = temp.dot(iCov*temp)/(m->gs2);
}

void Vert::findDepth() {
  Vec3d center(0,0,-1);
  depth = x.distTo(center) - 0.8;
}

void Vert::refreshLocale() {
  for (int j=0;j<nf;j++) {
    f[j]->findNormal();
    for (int k=0;k<3;k++) f[j]->e[k]->flag = -1;
  }
  for (int j=0;j<nf;j++) for (int k=0;k<3;k++) {
    if (f[j]->e[k]->flag != i) {
      f[j]->e[k]->findPotential();
      f[j]->e[k]->flag = i;
    }
  }
  for (int j=0;j<nf;j++) for (int k=0;k<3;k++) f[j]->e[k]->flag = -1;
}

void Vert::findLocalEdgePotential() {
  psiEdge = 0;

  for (int j=0;j<nf;j++) {
    for (int k=0;k<3;k++) {
      if (f[j]->e[k]->flag != i) {
  psiEdge += f[j]->e[k]->psi;
  f[j]->e[k]->flag = i;
      }
    }
  }
  psiLocal = psi + psiEdge;
}

void Vert::findGradient() {
  gradPsi0 = gradPsi;
  normGradPsi0 = normGradPsi;
  gradPsi = (2/m->gs2)*(x - x0);
  gradPsi *= iCov;
  for (int j=0;j<nf;j++) {
    for (int k=0;k<3;k++) {
      if (f[j]->e[k]->flag != i) {
  gradPsi += f[j]->e[k]->gradient(*this);
  f[j]->e[k]->flag = i;
      }
    }
  }
  normGradPsi = ~gradPsi;
}

void Vert::findSearchDir(float beta) {
  dir = dir0;
  dir *= beta;
  dir -= gradPsi;
}


void Vert::move(float globalStep) {
  if (!bound || 1) {
    x = xr;
    x += globalStep*dir;
  }
}

void Vert::move() {
  if (!bound || 1) {
    x = xr;
    x += step*dir;
  }
}

void Vert::flatten() {
  Vec3d delta = 0.0f;
  for (int j=0;j<nf;j++) {
    //delta += f[j]->n*(xr - f[j]->x).dot(f[j]->n);
    delta += (xr - f[j]->x);
  }
  if (nf > 0) delta /= nf;
  if (bound && 0)
    x -= n*delta.dot(n);
  else
    x -= delta;
}

// Equality tests
bool Vert::operator == (const Vert &vp) const {
  return (this == &vp);
}
bool Vert::operator != (const Vert &vp) const {
  return (this != &vp);
}

// Define << operator to handle output to screen
ostream& operator << (ostream& os, const Vert &v) {
  return os<<"v"<<v.i<<": "<<v.x<<endl<<flush;
}

} // namespace bmtk
